﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading.Tasks;
using MathNet.Numerics.Distributions;

namespace CustomerBehaviour
{
    class ModelObject
    {
        private readonly Random _localRand;
        private readonly double _premiumAlpha0;
        private readonly double _premiumAlpha1;
        private readonly double _premiumAlpha2;
        private readonly double _premiumStdDev;
        private double _marketPremium;
        private double _oldMarketPremium;
        private readonly double _startOldPremium;
        private readonly double _startOldOldPremium;

        private readonly List<Insurer> insurerList;
        private readonly List<Customer> customerList;
        private readonly int insCount;
        private readonly int custCount;
        private int renewCount;
        private readonly int _meanNetworkDegree;
        private readonly double _networkRewireProb;

        private readonly double[] newPremiumList;
        private readonly double[] serviceSpendPerCustomerList;
        private readonly int[] serviceExperienceList;
        private readonly int[] serviceCountList;
        private readonly double[] insProfitList;

        private readonly Outputs outputs;
        private readonly Outputs shareOutputs;
        private readonly CustomerOutputs customerOutputs;
        private double[] oldPremCost;
        private double[] oldLocCost;
        private double[] oldExpCost;
        private double[] oldWOMCost;

        internal ModelObject(int startSeed, string outputFolder, double oldPremium, double oldOldPremium, double premiumAlpha0, double premiumAlpha1, double premiumAlpha2, double marketPremiumStdDev, double renewPremiumMarkup, int renewMarkupMaxTerm, int numCustomers, int numInsurers, int meanNetworkDegree, double networkRewireProb, double maxServiceSpendPerCustomer, ISpendStrategy[] insurerSpendStrategyList, double distanceMultiplier, double serviceMultiplier, double memoryFactor, double influenceFactor, double renewAlpha, double renewBeta, double freqLambda, double sevAlpha, double sevBeta, double transmissionProb, bool writeDistributions)
        {
            //Set up local variables
            _localRand = new Random(startSeed);
            _premiumAlpha0 = premiumAlpha0;
            _premiumAlpha1 = premiumAlpha1;
            _premiumAlpha2 = premiumAlpha2;
            _premiumStdDev = marketPremiumStdDev;
            _marketPremium = oldPremium;
            _oldMarketPremium = oldOldPremium;
            _startOldPremium = oldPremium;
            _startOldOldPremium = oldOldPremium;

            //Initialise arrays
            insCount = numInsurers;
            newPremiumList = new double[insCount];
            serviceSpendPerCustomerList = new double[insCount];
            insProfitList = new double[insCount];
            custCount = numCustomers;
            serviceExperienceList = new int[custCount];
            serviceCountList = new int[custCount];
            oldPremCost = new double[custCount];
            oldLocCost = new double[custCount];
            oldExpCost = new double[custCount];
            oldWOMCost = new double[custCount];

            //Create list of insurer locations
            insurerList = new List<Insurer>();
            double insLocationStep = 2.0 / insCount;
            double[] insurerLocationList = new double[numInsurers];
            for(int i=0; i<insCount; i++)
            {
                insurerLocationList[i] = i * insLocationStep;
            }

            //Keep track of starting insurer choices to get old sold counts and old premium correct for first step renewal rates
            customerList = new List<Customer>();
            int[] oldSoldCounts = new int[insCount];
            int[][] oldSoldByIndex = new int[insCount][];
            for(int i=0;i< insCount;i++)
            {
                oldSoldByIndex[i] = new int[custCount];
            }
            double custLocationStep = 2.0 / custCount;
            
            int[] custInsurerChoices = new int[custCount];
            for (int i=0; i<custCount; i++)
            {
                int oldInsurer = _localRand.Next(0, numInsurers);
                custInsurerChoices[i] = oldInsurer;
                oldSoldCounts[oldInsurer]++;
                oldSoldByIndex[oldInsurer][i] = 1;
            }
            //Calculate old premium by insurer
            double[] insOldPremium = new double[insCount];
            for(int i=0; i< insCount; i++)
            {
                serviceSpendPerCustomerList[i] = insurerSpendStrategyList[i].GetOldServiceSpendPerCustomer();
                insOldPremium[i] = oldPremium + serviceSpendPerCustomerList[i];
            }
            //Create customer agents
            for (int i = 0; i < custCount; i++)
            {
                Customer cust = new Customer(i, i * custLocationStep, _localRand, insurerLocationList, custInsurerChoices[i], insOldPremium[custInsurerChoices[i]], distanceMultiplier, serviceMultiplier, memoryFactor, influenceFactor, renewAlpha, renewBeta, freqLambda, sevAlpha, sevBeta, transmissionProb);
                customerList.Add(cust);
                oldPremCost[i] = cust.premiumCost;
                oldLocCost[i] = cust.locationCost;
                oldExpCost[i] = cust.serviceCost;
                oldWOMCost[i] = cust.womCost;
            }

            //Create insurers and give them count of existing customers
            for (int i = 0; i < insCount; i++)
            {
                insurerList.Add(new Insurer(_localRand, i, outputFolder, i * insLocationStep, renewPremiumMarkup, renewMarkupMaxTerm, maxServiceSpendPerCustomer, insurerSpendStrategyList[i], oldSoldCounts[i],oldSoldByIndex[i]));
            }

            //Generate customer network
            _meanNetworkDegree = meanNetworkDegree;
            _networkRewireProb = networkRewireProb;
            GenerateNewNetwork();

            //Initialise csv writers
            shareOutputs = new Outputs(outputFolder);
            int outCount = Math.Min(10, insCount);
            var sortedList = insurerList.OrderByDescending(x => x.soldCount).ToList();
            for (int i = 0; i < insCount; i++)
            {
                if (i < outCount)
                {
                    shareOutputs.AddOutput("Share_" + (i + 1));
                }
                sortedList[i].SetNewRank(i);
            }
            shareOutputs.AddOutput("Top 10 Swap Count");

            outputs = new Outputs(outputFolder);
            outputs.AddOutput("Market Premium");
            outputs.AddOutput("Market Profit");
            outputs.AddOutput("Market Renewal Rate");
            outputs.AddOutput("Market Average Opinion");
            outputs.AddOutput("Customer_All_Average Start Premium Cost");
            outputs.AddOutput("Customer_All_Average Start Location Cost");
            outputs.AddOutput("Customer_All_Average Start Experience Cost");
            outputs.AddOutput("Customer_All_Average Start WOM Cost");
            outputs.AddOutput("Customer_All_Average Renewal Premium Cost Change");
            outputs.AddOutput("Customer_All_Average Renewal Location Cost Change");
            outputs.AddOutput("Customer_All_Average Renewal Experience Cost Change");
            outputs.AddOutput("Customer_All_Average Renewal WOM Cost Change");
            outputs.AddOutput("Customer_Renewals_Average Start Premium Cost");
            outputs.AddOutput("Customer_Renewals_Average Start Location Cost");
            outputs.AddOutput("Customer_Renewals_Average Start Experience Cost");
            outputs.AddOutput("Customer_Renewals_Average Start WOM Cost");
            outputs.AddOutput("Customer_Renewals_Average Renewal Premium Cost Change");
            outputs.AddOutput("Customer_Renewals_Average Renewal Location Cost Change");
            outputs.AddOutput("Customer_Renewals_Average Renewal Experience Cost Change");
            outputs.AddOutput("Customer_Renewals_Average Renewal WOM Cost Change");
            outputs.AddOutput("Customer_Switched_Average Start Premium Cost");
            outputs.AddOutput("Customer_Switched_Average Start Location Cost");
            outputs.AddOutput("Customer_Switched_Average Start Experience Cost");
            outputs.AddOutput("Customer_Switched_Average Start WOM Cost");
            outputs.AddOutput("Customer_Switched_Average Renewal Premium Cost Change");
            outputs.AddOutput("Customer_Switched_Average Renewal Location Cost Change");
            outputs.AddOutput("Customer_Switched_Average Renewal Experience Cost Change");
            outputs.AddOutput("Customer_Switched_Average Renewal WOM Cost Change");
            customerOutputs = new CustomerOutputs(outputFolder, insCount, custCount, writeDistributions);
            customerOutputs.AddOutput("Customer Experience");
            customerOutputs.AddOutput("Customer WOM");
            customerOutputs.AddOutput("Customer Opinion");
        }
        private void GenerateNewNetwork()
        {
            List<int>[] linkIndexLists = NetworkBuilder.BuildSmallWorldNetwork(_localRand, custCount, _meanNetworkDegree, _networkRewireProb);
            for (int i = 0; i < custCount; i++)
            {
                List<Customer> friendsList = new List<Customer>();
                List<int> linkIndexList = linkIndexLists[i];
                for (int j = 0; j < linkIndexList.Count; j++)
                {
                    friendsList.Add(customerList[linkIndexList[j]]);
                }
                customerList[i].SetFriendsList(friendsList);
            }
        }
        internal double[] RunTimeStep()
        {
            //Insurers choose how much to spend on customer service
            double[] newServiceSpendList = new double[insCount];
            System.Threading.Tasks.ParallelOptions opt = new System.Threading.Tasks.ParallelOptions();
            IEnumerable<int> iCustList = Enumerable.Range(0, custCount);
            IEnumerable<int> iInsList = Enumerable.Range(0, insCount);

            Parallel.ForEach<int>(iInsList, opt, ins =>
            {
                newServiceSpendList[ins] = insurerList[ins].CalculateSpending(serviceSpendPerCustomerList, insProfitList);
            });

            //set market premium and get insurer premium
            double newMarketPremium=_premiumAlpha0+_premiumAlpha1*_marketPremium+_premiumAlpha2*_oldMarketPremium+ Normal.Sample(_localRand, 0, _premiumStdDev);
            _oldMarketPremium = _marketPremium;
            _marketPremium = newMarketPremium;
            Parallel.ForEach<int>(iInsList, opt, ins =>
            {
                serviceSpendPerCustomerList[ins] = newServiceSpendList[ins];
                newPremiumList[ins] = insurerList[ins].CalculatePremium(_marketPremium);
            });

            //customers decide whether to renew or choose from new quotes
            int[] custRenew = new int[custCount];
            double[] newPremCost = new double[custCount];
            double[] newLocCost = new double[custCount];
            double[] newExpCost = new double[custCount];
            double[] newWOMCost = new double[custCount];
            double[] newRenewPremCost = new double[custCount];
            double[] newRenewLocCost = new double[custCount];
            double[] newRenewExpCost = new double[custCount];
            double[] newRenewWOMCost = new double[custCount];
            Parallel.ForEach<int>(iCustList, opt, custIndex =>
            {
                Customer cust = customerList[custIndex];
                int oldChoice = cust.chosenInsurer;
                double renewPrice = insurerList[oldChoice].GetRenewalPrice(custIndex);
                bool renew = cust.Renew(renewPrice);
                int newChoice= renew ? oldChoice : cust.ChooseNew(newPremiumList);
                if (newChoice == oldChoice) { custRenew[custIndex] = 1; }
                newPremCost[custIndex] = cust.premiumCost;
                newLocCost[custIndex] = cust.locationCost;
                newExpCost[custIndex] = cust.serviceCost;
                newWOMCost[custIndex] = cust.womCost;
                newRenewPremCost[custIndex] = cust.renewPremiumCost;
                newRenewLocCost[custIndex] = cust.renewLocationCost;
                newRenewExpCost[custIndex] = cust.renewServiceCost;
                newRenewWOMCost[custIndex] = cust.renewWomCost;
            });
            renewCount = custRenew.Sum();
            for(int i=0;i< custCount;i++)
            {
                Customer cust = customerList[i];
                insurerList[cust.chosenInsurer].AddCustomer(i, cust.location);
            }

            //Generate losses. If loss->interaction with customer service, good or bad
            for(int i=0; i<custCount; i++)
            {
                Customer cust = customerList[i];
                Loss newLoss = cust.GetLoss();
                Insurer ins = insurerList[cust.chosenInsurer];
                ins.AddLoss(newLoss.sev);
                serviceCountList[i] = newLoss.freq;
                int totalExperience = 0;
                for (int j=0;j<newLoss.freq;j++)
                {
                    totalExperience += ins.CalculateServiceExperience();
                }
                serviceExperienceList[i] = totalExperience;
            }

            //Update customer information with new customer service experience and WOM
            List<double[]> custExperience = new List<double[]>();
            List<double[]> custWOM = new List<double[]>();
            Parallel.ForEach<int>(iCustList, opt, custIndex =>
            {
                Customer cust = customerList[custIndex];
                cust.UpdateService(serviceCountList[custIndex], serviceExperienceList[custIndex]);
                cust.UpdateWOM();
            });

            //Now WOM has happened based on existing opinion, update opinion with new info
            List<double[]> custOpinion = new List<double[]>();
            double avgOpinion = 0;
            for (int i = 0; i < custCount; i++)
            {
                Customer cust = customerList[i];
                cust.UpdateOpinion();
                custExperience.Add(cust.insServiceExperienceAvg);
                custWOM.Add(cust.insServiceWOMAvg);
                custOpinion.Add(cust.insServiceBonus);
                avgOpinion += cust.currentOpinion / (double)custCount;
            }
            //Update insurer profits
            for (int i=0; i< insCount; i++)
            {
                insProfitList[i] = insurerList[i].CalculateProfit();
            }
            //Get customer cost info
            double avgStartPremCost = 0;
            double avgStartLocCost = 0;
            double avgStartExpCost = 0;
            double avgStartWOMCost = 0;
            double avgRenewPremCost = 0;
            double avgRenewLocCost = 0;
            double avgRenewExpCost = 0;
            double avgRenewWOMCost = 0;
            double renewAvgStartPremCost = 0;
            double renewAvgStartLocCost = 0;
            double renewAvgStartExpCost = 0;
            double renewAvgStartWOMCost = 0;
            double renewAvgRenewPremCost = 0;
            double renewAvgRenewLocCost = 0;
            double renewAvgRenewExpCost = 0;
            double renewAvgRenewWOMCost = 0;
            double newAvgStartPremCost = 0;
            double newAvgStartLocCost = 0;
            double newAvgStartExpCost = 0;
            double newAvgStartWOMCost = 0;
            double newAvgRenewPremCost = 0;
            double newAvgRenewLocCost = 0;
            double newAvgRenewExpCost = 0;
            double newAvgRenewWOMCost = 0;
            int newCount = custCount - renewCount;
            for (int i=0;i<custCount; i++)
            {
                avgStartPremCost += oldPremCost[i] / custCount;
                avgStartLocCost += oldLocCost[i] / custCount;
                avgStartExpCost += oldExpCost[i] / custCount;
                avgStartWOMCost += oldWOMCost[i] / custCount;
                avgRenewPremCost += newRenewPremCost[i] / custCount;
                avgRenewLocCost += newRenewLocCost[i] / custCount;
                avgRenewExpCost += newRenewExpCost[i] / custCount;
                avgRenewWOMCost += newRenewWOMCost[i] / custCount;
                if(custRenew[i]>0)
                {
                    renewAvgStartPremCost += oldPremCost[i] /renewCount;
                    renewAvgStartLocCost += oldLocCost[i] / renewCount;
                    renewAvgStartExpCost += oldExpCost[i] / renewCount;
                    renewAvgStartWOMCost += oldWOMCost[i] / renewCount;
                    renewAvgRenewPremCost += newRenewPremCost[i] / renewCount;
                    renewAvgRenewLocCost += newRenewLocCost[i] / renewCount;
                    renewAvgRenewExpCost += newRenewExpCost[i] / renewCount;
                    renewAvgRenewWOMCost += newRenewWOMCost[i] / renewCount;
                }
                else
                {
                    newAvgStartPremCost += oldPremCost[i] / newCount;
                    newAvgStartLocCost += oldLocCost[i] / newCount;
                    newAvgStartExpCost += oldExpCost[i] / newCount;
                    newAvgStartWOMCost += oldWOMCost[i] / newCount;
                    newAvgRenewPremCost += newRenewPremCost[i] / newCount;
                    newAvgRenewLocCost += newRenewLocCost[i] / newCount;
                    newAvgRenewExpCost += newRenewExpCost[i] / newCount;
                    newAvgRenewWOMCost += newRenewWOMCost[i] / newCount;
                }
            }
            oldPremCost = newPremCost;
            oldLocCost = newLocCost;
            oldExpCost = newExpCost;
            oldWOMCost = newWOMCost;

            //Write outputs
            var sortedList = insurerList.OrderByDescending(x => x.soldCount).ToList();
            int outCount = Math.Min(10, insCount);
            int swapCount = 0;
            var outValues = new double[outCount + 1];
            for (int i = 0; i < insCount; i++)
            {
                if (i < outCount)
                {
                    outValues[i] = sortedList[i].soldCount / (double)custCount;
                    swapCount += sortedList[i].SetNewRank(i);
                }
                else
                {
                    sortedList[i].SetNewRank(i);
                }
            }
            outValues[outCount] = swapCount;
            shareOutputs.SetValues(outValues);
            shareOutputs.WriteOutputs();
            outputs.SetValues(new double[] { _marketPremium, insProfitList.Sum(), renewCount / (double)custCount, avgOpinion,
            avgStartPremCost, avgStartLocCost, avgStartExpCost, avgStartWOMCost, avgRenewPremCost-avgStartPremCost, avgRenewLocCost-avgStartLocCost, avgRenewExpCost-avgStartExpCost, avgRenewWOMCost-avgStartWOMCost,
            renewAvgStartPremCost, renewAvgStartLocCost, renewAvgStartExpCost, renewAvgStartWOMCost, renewAvgRenewPremCost-renewAvgStartPremCost, renewAvgRenewLocCost-renewAvgStartLocCost, renewAvgRenewExpCost-renewAvgStartExpCost, renewAvgRenewWOMCost-renewAvgStartWOMCost,
            newAvgStartPremCost, newAvgStartLocCost, newAvgStartExpCost, newAvgStartWOMCost, newAvgRenewPremCost-newAvgStartPremCost, newAvgRenewLocCost-newAvgStartLocCost, newAvgRenewExpCost-newAvgStartExpCost, newAvgRenewWOMCost-newAvgStartWOMCost});
            outputs.WriteOutputs();
            customerOutputs.SetValues(new List<List<double[]>> { custExperience, custWOM, custOpinion });
            customerOutputs.WriteOutputs();

            for (int i = 0; i < insCount; i++)
            {
                insurerList[i].NextTimestep();
            }
            return insProfitList;
        }
        internal void NextSim(ISpendStrategy[] insurerSpendStrategyList)
        {
            //Reset values and output objects
            _marketPremium = _startOldPremium;
            _oldMarketPremium = _startOldOldPremium;
            shareOutputs.NextSim();
            outputs.NextSim();
            customerOutputs.NextSim();

            //Need to re-initialise starting insurers
            //Keep track of starting insurer choices to get old sold counts and old premium correct for first step renewal rates
            int[] oldSoldCounts = new int[insCount];
            int[][] oldSoldByIndex = new int[insCount][];
            int[] custInsurerChoices = new int[custCount];
            for (int i = 0; i < insCount; i++)
            {
                oldSoldByIndex[i] = new int[custCount];
            }
            for (int i = 0; i < custCount; i++)
            {
                int oldInsurer = _localRand.Next(0, insCount);
                custInsurerChoices[i] = oldInsurer;
                oldSoldCounts[oldInsurer]++;
                oldSoldByIndex[oldInsurer][i] = 1;
            }
            //Calculate old premium by insurer and set up customers for next sim
            double[] insOldPremium = new double[insCount];
            for (int i = 0; i < insCount; i++)
            {
                serviceSpendPerCustomerList[i]= insurerSpendStrategyList[i].GetOldServiceSpendPerCustomer();
                insOldPremium[i] = _startOldPremium + serviceSpendPerCustomerList[i];
            }
            for (int i=0;i< custCount;i++)
            {
                Customer cust = customerList[i];
                customerList[i].NextSim(custInsurerChoices[i], insOldPremium[custInsurerChoices[i]]);
                oldPremCost[i] = cust.premiumCost;
                oldLocCost[i] = cust.locationCost;
                oldExpCost[i] = cust.serviceCost;
                oldWOMCost[i] = cust.womCost;
            }
            //Now set up insurers for next sim
            for (int i=0;i< insCount;i++)
            {
                insurerList[i].NextSim(oldSoldCounts[i], oldSoldByIndex[i], insurerSpendStrategyList[i]);
            }
            GenerateNewNetwork();
            var sortedList = insurerList.OrderByDescending(x => x.soldCount).ToList();
            for (int i = 0; i < insCount; i++)
            {
                sortedList[i].SetNewRank(i);
            }
        }
        internal void Close()
        {
            shareOutputs.Close();
            outputs.Close();
            customerOutputs.Close();
            for (int i = 0; i < insCount; i++)
            {
                insurerList[i].Close();
            }     
        }
    }
}