﻿using System;
using System.Collections.Generic;

namespace CustomerBehaviour
{
    static class NetworkBuilder
    {
        internal static List<int>[] BuildSmallWorldNetwork(Random localRand, int numAgents, int meanDegree, double rewireProb)
        {
            //First randomly shuffle index numbers
            int[] shuffledIndices = new int[numAgents];
            for(int i=0;i<numAgents;i++)
            {
                shuffledIndices[i] = i;
            }
            Shuffle(shuffledIndices, localRand);

            //Now draw initial list of links
            //Initialise from and to arrays for ach agent showing list of links to add
            int numRightLinks= meanDegree / 2;
            List<int>[] linkFromLists = new List<int>[numAgents];
            List<int>[] linkToLists = new List<int>[numAgents];
            for (int i = 0; i < numAgents; i++)
            {
                linkFromLists[i] = new List<int>();
                linkToLists[i] = new List<int>();
            }
            //Now add links from each agent to immediate right hand neighbours
            for(int i=0;i< numAgents;i++)
            {
                for(int j=0;j< numRightLinks;j++)
                {
                    int fromIndex = shuffledIndices[i];
                    int toIndex = i + j + 1 >= numAgents ? shuffledIndices[i + j + 1 - numAgents] : shuffledIndices[i + j + 1];
                    linkFromLists[fromIndex].Add(toIndex);
                    linkToLists[toIndex].Add(fromIndex);
                }
            }

            //Go through list and apply chance of rewiring (if rewiring is possible!)
            if(numRightLinks<numAgents/2)
            {
                for(int i=0; i< numAgents; i++)
                {
                    for (int j = 0; j < numRightLinks; j++)
                    {
                        if (localRand.NextDouble() < rewireProb)
                        {
                            //Get new link destination
                            int oldTo = linkFromLists[i][j];
                            int newTo = GetNewNeighbour(localRand, numAgents, i, oldTo, linkFromLists[i], linkToLists[i]);
                            linkFromLists[i][j] = newTo;
                            //Now remember to fix it in the 'to' lists!
                            linkToLists[oldTo].Remove(i);
                            linkToLists[newTo].Add(i);
                        }
                    }
                }
            }

            //Finally, add all links together to send the final (bi-directional) return array
            //Add index of link partner to list for agents on both sides of link
            for(int i=0;i< numAgents;i++)
            {
                for(int j=0;j<linkToLists[i].Count;j++)
                {
                    linkFromLists[i].Add(linkToLists[i][j]);
                }
            }
            return linkFromLists;
        }

        //Fisher-Yates shuffle
        public static void Shuffle<T>(T[] array, Random localRand)
        {
            int n = array.Length;
            while (n>1)
            {
                n--;
                int k = localRand.Next(n+1);
                T t = array[k];
                array[k] = array[n];
                array[n] = t;
            }
        }

        //Finding a new target for rewired link
        private static int GetNewNeighbour(Random localRand, int numAgents, int fromLink, int toLink, List<int> linkFromList, List<int> linkToList)
        {
            bool linkCheck = false;
            int newToLink = toLink;
            while (linkCheck == false)
            {
                //Generate new random. Must exclude self-link
                newToLink = localRand.Next(numAgents - 1);
                if (newToLink >= fromLink) { newToLink++; }
                //Now check that this link does not already exist
                linkCheck = true;
                {
                    for(int i=0;i<linkFromList.Count;i++)
                    {
                        if( (linkFromList[i]==newToLink))
                        {
                            linkCheck = false;
                        }
                    }
                    for (int i = 0; i < linkToList.Count; i++)
                    {
                        if ((linkToList[i] == newToLink))
                        {
                            linkCheck = false;
                        }
                    }
                }
            }
            return newToLink;
        }
    }
}
